#include "kanji_union.h"

extern FinalEffect *finalEffect;

NORMAL_SUBSTITUTE data[DATA_ARRAY_LIMIT];

int kanjiCode(std::string & line) {
    using namespace std;

    bool kanji = false;
    char *buffer;
    char *p;
    int before;
    int k, n;
    int size = line.length();
    int size_1 = size + 1;
    unsigned char uc;

    try {
        buffer = new char[size_1];
        strcpy(buffer, line.c_str() );

        for(int i = 0; i < size; ++i) {
            uc = (unsigned char)buffer[i];
            k = (int)uc;

            if(kanji) {
                p = & buffer[before];
                n = inverted(p);
                break;
            }

            if(0x81 <= k && k <= 0xFC) {
                if (KANA_KUTEN <= k && k <= KANA_HANDAKUTEN) {
                    n = k;
                    break;
                } else {
                    before = i;
                    kanji = true;
                }

            } else {
                n = k;
                break;
            }

        }

        delete[] buffer;
    } catch(...) {
        errorHandler("new char[] error-7:", size_1);
//        exit(1);
    }

    return n;
}

bool editRecord(std::string & normal, std::string & substitute, std::string & binaryName, bool no_searching, bool *whichTable) {
    using namespace std;

    char buffer[BLOCK_SIZE];
    char c[DOUBLE_KANJI];
    int buffer_size = BLOCK_SIZE;
    int from, into, k;
    int manager = finalEffect->encodingMethod();
    OPEN_CLOSE_FUNCTION *openClose = new OPEN_CLOSE_FUNCTION[SEARCH_COUNT];
    stringstream os;
    unsigned char uc;

    openClose[0] = binary_file_in_installed_directory;
    openClose[1] = binary_file_in_current_directory;
    openClose[2] = binary_file_in_path_directory;
    openClose[3] = binary_file_in_specified_directory;

    for(int i = 0; i < SEARCH_COUNT; ++i) {
        if( openClose[i](buffer, buffer_size) ) {
            if(manager == TO_EUC) {
                if(*whichTable) {
                    messagesText(buffer, "is primarily accessible to SideAce.");
                    *whichTable = false;
                }

            }

            break;
        } else {
            switch(i) {
            case 0:
                if(no_searching) {
                    errorHandler("openClose[", i, "]() error-3:", SJIS_EUC_BIN);
                    delete[] openClose;
                    return false;
                }

                break;
            case 3:
                errorHandler("openClose[", i, "]() error-4:", SJIS_EUC_BIN);
                delete[] openClose;
                return false;
            default:
                break;
            }

        }

    }

    SparseEUC *euc = new SparseEUC();
    euc->clear();
    if(! euc->setting(no_searching, whichTable) ) {
        errorHandler("editRecord() error:", "euc->setting(no_searching, whichTable)");
        delete euc;
        delete[] openClose;
        return false;
    }

    from = kanjiCode(normal);
    into = kanjiCode(substitute);
    infoText(from, RIGHT_ARROW, into);

    ifstream binaryFile(buffer);
    for(int i = 1; i <= RECORD_COUNT; ++i) {
        binaryFile.read(c, DOUBLE_KANJI);
        if(i < FIRST_KANJI) {
            uc = (unsigned char)c[0];
            k = (int)uc;
        } else {
            k = inverted(c);
        }

        os << c[0] << c[1];

        if(k == from) {
            os << euc->working(into, 0) << euc->working(into, 1);
        } else {
            os << c[SJIS_KANJI_SIZE] << c[ SJIS_KANJI_SIZE + 1 ];
        }

    }

    binaryFile.close();
    ofstream editedFile( binaryName.c_str(), ios::out | ios::binary | ios::trunc );
    editedFile << os.str();
    delete euc;
    delete[] openClose;
    return true;
}

bool editManyRecords(std::string & text, std::string & binaryName, bool no_searching, bool *whichTable) {
    using namespace std;

    bool checked = false;
    char buffer[BLOCK_SIZE];
    char c[DOUBLE_KANJI];
    int buffer_size = BLOCK_SIZE;
    int from, into, k;
    int m = 0;
    int manager = finalEffect->encodingMethod();
    int size = 0;
    int used = 0;
    OPEN_CLOSE_FUNCTION *openClose = new OPEN_CLOSE_FUNCTION[SEARCH_COUNT];
    stringstream os;
    unsigned char uc;

    openClose[0] = binary_file_in_installed_directory;
    openClose[1] = binary_file_in_current_directory;
    openClose[2] = binary_file_in_path_directory;
    openClose[3] = binary_file_in_specified_directory;

    for(int i = 0; i < SEARCH_COUNT; ++i) {
        if( openClose[i](buffer, buffer_size) ) {
            if(manager == TO_EUC) {
                if(*whichTable) {
                    messagesText(buffer, "is primarily accessible to SideAce.");
                    *whichTable = false;
                }

            }

            break;
        } else {
            switch(i) {
            case 0:
                if(no_searching) {
                    errorHandler("openClose[", i, "]() error-5:", SJIS_EUC_BIN);
                    delete[] openClose;
                    return false;
                }

                break;
            case 3:
                errorHandler("openClose[", i, "]() error-6:", SJIS_EUC_BIN);
                delete[] openClose;
                return false;
            default:
                break;
            }

        }

    }

    SparseEUC *euc = new SparseEUC();
    euc->clear();
    if(! euc->setting(no_searching, whichTable) ) {
        errorHandler("editManyRecords() error:", "euc->setting(no_searching, whichTable)");
        delete euc;
        delete[] openClose;
        return false;
    }

    ifstream sjisFile( text.c_str() );
    if( sjisFile.fail() ) {
        errorHandler("ifstream sjisFile() error:", text);
        delete euc;
        delete[] openClose;
        return false;
    }

    char record[BLOCK_SIZE];
    string *d = new string("");
    while( sjisFile.getline(record, BLOCK_SIZE) ) {
        delete d;
        d = new string(record);
        size = d->length();
        if(size == DUAL_KANJI) {
            data[m].please_check = true;
            data[m].normal = d->substr(0, 2);
            data[m].substitute = d->substr(SJIS_KANJI_SIZE, 2);
        } else {
            data[m].please_check = true;
            data[m].normal = d->substr(0, 1); // katakana, for example 
            data[m].substitute = d->substr(KATAKANA_SIZE, 1);
        }

        used = m;
        m += 1;
        if(m >= DATA_ARRAY_LIMIT) {
            break;
        }

    }

    ifstream binaryFile(buffer);
    for(int i = 1; i <= RECORD_COUNT; ++i) {
        binaryFile.read(c, DOUBLE_KANJI);
        if(i < FIRST_KANJI) {
            uc = (unsigned char)c[0];
            k = (int)uc;
        } else {
            k = inverted(c);
        }

        os << c[0] << c[1];

        for(int m = 0; m <= used; ++m) {
            if(data[m].please_check) {
                from = kanjiCode(data[m].normal);
                if(k == from) {
                    into = kanjiCode(data[m].substitute);
                    checked = true;
                    data[m].please_check = false;
                    break;
                } else {
                    checked = false;
                }

            } else {
                checked = false;
            }

        }

        if(checked) {
            os << euc->working(into, 0) << euc->working(into, 1);
        } else {
            os << c[SJIS_KANJI_SIZE] << c[ SJIS_KANJI_SIZE + 1 ];
        }

    }

    binaryFile.close();
    ofstream editedFile( binaryName.c_str(), ios::out | ios::binary | ios::trunc );
    editedFile << os.str();
    delete d;
    delete euc;
    delete[] openClose;
    return true;
}

