#include "kanji_union.h"

KANJI kanji_union;
KANJI *double_byte_character = &kanji_union;

namespace {
    const int READING = false;
    const int WRITING = true;
}

bool openClose(char *file_name, bool write) {
    using namespace std;

    bool b = true;
    if(write) {
        ofstream out(file_name);
        if( out.fail() ) {
            b = false;
            strcpy(file_name, SIDE_ACE_TEXT);
        }

    } else {
        ifstream in(file_name);
        if( in.fail() ) {
            b = false;
            errorHandler("ifstream in() error:", file_name);
            exit(1);
        }

    }

    return b;
}

int inverted(char *c) {
    char *p = (char *)double_byte_character->sequence;

// Copying in reverse order
    p[0] = c[1];
    p[1] = c[0];
    return double_byte_character->code[0];
}

int Regulate::readingAnotherFileToProduceEUC(char *new_name) {
    using namespace std;

    bool kanji = false;
    char *p;
    char a[3] = {0, 0, 0};
    char newLine[3] = {0x0A, 0, 0};
    int c;
    int exit_code = 0;
    stringstream *os = new stringstream();
    unsigned char uc;
    unsigned char firstByte = 0x40;

    if(! openClose(file_name, READING) ) {
        cerr << "openClose() error-1: " << file_name << endl;
        return ERROR_CODE;
    }

    if(! openClose(new_name, WRITING) ) {
        cerr << "openClose() error-2: " << new_name << endl;
        return ERROR_CODE;
    }

    FILE *file = fopen(file_name, "rt");
    ofstream eucFile(new_name, ios::out | ios::binary| ios::trunc );
    SparseEUC *euc = new SparseEUC();
    euc->clear();
    if(! euc->setting(no_searching, &whichTable) ) {
        errorHandler("readingAnotherFileToProduceEUC() error:", "euc->setting(no_searching, &whichTable)");
        return 1;
    }

    if(carriage_return) {
        newLine[0] = 0x0D;
        newLine[1] = 0x0A;
    }

    for(;;) {
        p = fgets(buffer, BUFFER_SIZE, file);
        if(p == NULL) {
            break;
        } else {
            p = strchr(buffer, '\n');
            if(p != NULL) {
                *p = '\0';
            }

        }

        delete os;
        os = new stringstream();
        for(int i = 0; buffer[i]; ++i) {
            uc = (unsigned char)buffer[i];
            c = (int)uc;

            if(kanji) {
                a[0] = (char)firstByte;
                a[1] = buffer[i];
                *os << euc->working( inverted(a) );
                kanji = false;
                continue;
            }

            if(0x81 <= c && c<= 0xFC) {
                if(KANA_KUTEN <= c && c <= KANA_HANDAKUTEN) {
                    *os << euc->working(c);
                    kanji = false;
                } else {
                    firstByte = uc;
                    kanji = true;
                }

            } else {
                *os << uc;
                kanji = false;
            }

        }

        eucFile << os->str() << newLine;
    }

    fclose(file);
    delete os;
    delete euc;
    return exit_code;
}

int Regulate::readingToProduceEUC(char *new_name) {
    using namespace std;

    bool kanji = false;
    char a[3] = {0, 0, 0};
    char newLine[3] = {0x0A, 0, 0};
    int c;
    int exit_code = 0;
    string s;
    stringstream *os = new stringstream();
    unsigned char uc;
    unsigned char firstByte = 0x40;

    if(! openClose(new_name, WRITING) ) {
        errorHandler("openClose() error-2:", new_name);
        return ERROR_CODE;
    }

    ofstream eucFile(new_name, ios::out | ios::binary| ios::trunc );
    SparseEUC *euc = new SparseEUC();
    euc->clear();
    if(! euc->setting(no_searching, &whichTable) ) {
        errorHandler("readingToProduceEUC() error:", "euc->setting(no_searching, &whichTable)");
        return ERROR_CODE;
    }

    if(carriage_return) {
        newLine[0] = 0x0D;
        newLine[1] = 0x0A;
    }

    for(ITEM::iterator i = v.begin(); i != v.end(); ++i) {
        s = *i;

        delete os;
        os = new stringstream();

        strcpy(buffer, s.c_str() );
        for(int i = 0; buffer[i]; ++i) {
            uc = (unsigned char)buffer[i];
            c = (int)uc;

            if(kanji) {
                a[0] = (char)firstByte;
                a[1] = buffer[i];
                *os << euc->working( inverted(a) );
                kanji = false;
                continue;
            }

            if(0x81 <= c && c<= 0xFC) {
                if(KANA_KUTEN <= c && c <= KANA_HANDAKUTEN) {
                    *os << euc->working(c);
                    kanji = false;
                } else {
                    firstByte = uc;
                    kanji = true;
                }

            } else {
                *os << uc;
                kanji = false;
            }

        }

        eucFile << os->str() << newLine;
    }

    delete os;
    delete euc;
    return exit_code;
}

int Regulate::readingToProduceUTF8(char *new_name) {
    using namespace std;

    bool kanji = false;
    char a[3] = {0, 0, 0};
    char newLine[3] = {0x0A, 0, 0};
    int c;
    int exit_code = 0;
    string s;
    stringstream *os = new stringstream();
    unsigned char uc;
    unsigned char firstByte = 0x40;

    if(! openClose(new_name, WRITING) ) {
        errorHandler("openClose() error-2:", new_name);
        return ERROR_CODE;
    }

    ofstream utf8File(new_name, ios::out | ios::binary| ios::trunc );
    SparseUTF8 *utf8 = new SparseUTF8();
    utf8->clear();
    if(! utf8->setting(no_searching, &whichTable) ) {
        errorHandler("readingToProduceUTF8() error:", "utf8->setting(no_searching, &whichTable)");
        return ERROR_CODE;
    }

    if(carriage_return) {
        newLine[0] = 0x0D;
        newLine[1] = 0x0A;
    }

    for(ITEM::iterator i = v.begin(); i != v.end(); ++i) {
        s = *i;

        delete os;
        os = new stringstream();

        strcpy(buffer, s.c_str() );
        for(int i = 0; buffer[i]; ++i) {
            uc = (unsigned char)buffer[i];
            c = (int)uc;

            if(kanji) {
                a[0] = (char)firstByte;
                a[1] = buffer[i];
                *os << utf8->working( inverted(a) );
                kanji = false;
                continue;
            }

            if(0x81 <= c && c<= 0xFC) {
                if(KANA_KUTEN <= c && c <= KANA_HANDAKUTEN) {
                    *os << utf8->working(c);
                    kanji = false;
                } else {
                    firstByte = uc;
                    kanji = true;
                }

            } else {
                *os << uc;
                kanji = false;
            }

        }

        utf8File << os->str() << newLine;
    }

    delete os;
    delete utf8;
    return exit_code;
}

int produceBinaryFile(const char *name, bool no_removing) {
    using namespace std;

    char c[BLOCK_SIZE];
    FILE *ancient = fopen(name, "rb");
    int exit_code = 0;
    int size;
    FILE *temporary = fopen(TEMPORARY_BINARY, "wb");
    if(temporary == NULL) {
        cerr << "fopen() error-1: " << TEMPORARY_BINARY << endl;
        exit(1);
    }

    if(ancient == NULL) {
        exit_code = FILE_ERROR_CODE;
    } else {
        do {
            size = fread(c, 1, BLOCK_SIZE, ancient);
            fwrite(c, size, 1, temporary);
        } while(size == BLOCK_SIZE);

        fclose(temporary);
        fclose(ancient);

        if(! no_removing) {
            remove(name);
        }

    }

    return exit_code;
}

int renameBinaryFile(const char *name, bool no_removing) {
    char c[BLOCK_SIZE];
    int exit_code = 0;
    int size;
    FILE *temporary = fopen(TEMPORARY_BINARY, "rb");
    FILE *recent = fopen(name, "wb");
    if(recent == NULL) {
        recent = fopen(SIDE_ACE_TEXT, "wb");
    }

    if(temporary == NULL) {
        exit_code = FILE_ERROR_CODE;
    } else {
        do {
            size = fread(c, 1, BLOCK_SIZE, temporary);
            fwrite(c, size, 1, recent);
        } while(size == BLOCK_SIZE);

        fclose(recent);
        fclose(temporary);

        if(! no_removing) {
            remove(TEMPORARY_BINARY);
        }

    }

    return exit_code;
}

/*

1:  MS932~EUC-JP.bin   table                               
2:  S                 normal                              
3:                   substitute                          
4:  sortkCxx.txt       bars.txt                            
5:  -c3                corner                              
6:  -e3                manager                             
7:  -a                 IDENTICAL_TIME                      
8:  -b                 TEMPORARY_BINARY_FILE               
9:  -CRLF              -CRLF                               
10: -d                 DO_NOT_SEARCH_TABLE                 
11: -f                 RECOGNIZE_FULL_NAME                 
12: -i                 SKIPPING                            
13: -n                 NUMBER                              
14: -p                 SEE_PDF                             
15: -q                 NO_READING                          
16: -u                 NO_EDITING                          
17: -w                 WHICH_TABLE                         
18: -y                 DO_NOT_REMOVE_TEMPORARY_BINARY_FILE 
*/

void acquire_parameters_from_text(const char *parameters_text, char *binary_name, char *normal, char *substitute, char *text_of_substitutes_list, int *corner, int *manager,
 bool *identical_time, bool *produce_binary_file, bool *carriage_return, bool *no_searching, bool *full_name, bool *skip_control_characters, bool *line_begins_with_number, bool *document, bool *no_reading,
 bool *edit_table, bool *which_table, bool *no_removing) {
    char *p;
    char *stop;
    char c[C_SIZE];
    FILE *setting = fopen(parameters_text, "rt");
    int k;
    int line_number = 0;

    if(setting != NULL) {
        for(;;) {
            p = fgets(c, C_SIZE, setting);
            if(p == NULL) {
                break;
            } else {
                k = strcspn(c, "\n',|\t");
                c[k] = '\0';
            }

            line_number += 1;
            switch(line_number) {
            case 1:
                strcpy(binary_name, c); // MS932~EUC-JP.bin
                break;
            case 2:
                strcpy(normal, c); // S
                break;
            case 3:
                strcpy(substitute, c); // 
                break;
            case 4:
                strcpy(text_of_substitutes_list, c); // sortkCxx.txt
                break;
            case 5:
                p = & c[2];
                *corner = (int)strtol(p, &stop, 10);  // -c3
                break;
            case 6:
                p = & c[2];
                *manager = (int)strtol(p, &stop, 10); // -e3
                break;
            case 7:
                if(! strcmp(IDENTICAL_TIME, c) ) { // -a
                    *identical_time = true;
                } else if(! strcmp(ZERO, c) ) { // 0
                    *identical_time = false;
                } else {
                    *identical_time = true;
                }

                break;
            case 8:
                if(! strcmp(TEMPORARY_BINARY_FILE, c) ) { // -b
                    *produce_binary_file = true;
                } else if(! strcmp(ZERO, c) ) { // 0
                    *produce_binary_file = false;
                } else {
                    *produce_binary_file = true;
                }

                break;
            case 9:
                if(! strcmp(CRLF, c) ) { // -CRLF
                    *carriage_return = true;
                } else if(! strcmp(DO_NOT_REMOVE_0DH, c) ) { // -L-
                    *carriage_return = true;
                } else if(! strcmp(LF, c) ) { // -LF
                    *carriage_return = false;
                } else if(! strcmp(LINE_FEED, c) ) { // -L
                    *carriage_return = false;
                } else if(! strcmp(ZERO, c) ) { // 0
                    *carriage_return = false;
                } else {
                    *carriage_return = true;
                }

                break;
            case 10:
                if(! strcmp(DO_NOT_SEARCH_TABLE, c) ) { // -d
                    *no_searching = true;
                } else if(! strcmp(ZERO, c) ) { // 0
                    *no_searching = false;
                } else {
                    *no_searching = true;
                }

                break;
            case 11:
                if(! strcmp(RECOGNIZE_FULL_NAME, c) ) { // -f
                    *full_name = true;
                } else if(! strcmp(ZERO, c) ) { // 0
                    *full_name = false;
                } else {
                    *full_name = true;
                }

                break;
            case 12:
                if(! strcmp(SKIPPING, c) ) { // -i
                    *skip_control_characters = true;
                } else if(! strcmp(ZERO, c) ) { // 0
                    *skip_control_characters = false;
                } else {
                    *skip_control_characters = true;
                }

                break;
            case 13:
                if(! strcmp(NUMBER, c) ) { // -n
                    *line_begins_with_number = true;
                } else if(! strcmp(ZERO, c) ) { // 0
                    *line_begins_with_number = false;
                } else {
                    *line_begins_with_number = true;
                }

                break;
            case 14:
                if(! strcmp(SEE_PDF, c) ) { // -p
                    *document = true;
                } else if(! strcmp(ZERO, c) ) { // 0
                    *document = false;
                } else {
                    *document = true;
                }

                break;
            case 15:
                if(! strcmp(NO_READING, c) ) { // -q
                    *no_reading = true;
                } else if(! strcmp(ZERO, c) ) { // 0
                    *no_reading = false;
                } else {
                    *no_reading = true;
                }

                break;
            case 16:
                if(! strcmp(NO_EDITING, c) ) { // -u
                    *edit_table = false;
                } else if(! strcmp(ZERO, c) ) { // 0
                    *edit_table = true;
                } else {
                    *edit_table = false;
                }

                break;
            case 17:
                if(! strcmp(WHICH_TABLE, c) ) { // -w
                    *which_table = true;
                } else if(! strcmp(ZERO, c) ) { // 0
                    *which_table = false;
                } else {
                    *which_table = true;
                }

                break;
            case 18:
                if(! strcmp(WHICH_TABLE, c) ) { // -y
                    *no_removing = true;
                } else if(! strcmp(ZERO, c) ) { // 0
                    *no_removing = false;
                } else {
                    *no_removing = true;
                }

                break;
            default:
                break;
            } // switch(line_number)

        } // for(;;)

        fclose(setting);
    } // if(setting != NULL)

}
