#ifndef KANJI_UNION_H
#define KANJI_UNION_H

#ifdef __BORLANDC__
#pragma warn -8004
#pragma warn -8012
#endif

#define VERSION 1.08
#define ACE "ACE"
#define ARRAY_SIZE FINAL_SIZE_1 // Numbered.h
#define BELL 0x07
#define BELL_ASCII_KATAKANA_COUNT 224
#define BLOCK_SIZE 2560
#define BUFFER_SIZE 524288 // 512 KB
#define CRLF "-CRLF"
#define C_MS932_SJIS_BIN "C:\\SideAceVCL\\MS932~Shift_JIS.bin"
#define C_SIZE 256 // FILE_NAME_SIZE + 1
#define C_SJIS_EUC_BIN "C:\\SideAceVCL\\MS932~EUC-JP.bin"
#define C_SJIS_UTF8_BIN "C:\\SideAceVCL\\MS932~UTF-8.bin"
#define DATA_ARRAY_LIMIT 906
#define DO_NOT_REMOVE_0DH "-L-"
#define DO_NOT_REMOVE_TEMPORARY_BINARY_FILE "-y"
#define DO_NOT_SEARCH_TABLE "-d"
#define DOUBLE_KANJI 4 // SJIS_KANJI_SIZE + EUC_KANJI_SIZE
#define DOUBLE_KANJI_UTF8 5 // SJIS_KANJI_SIZE + UTF8_KANJI_SIZE
#define DUAL_KANJI 4 // SJIS_KANJI_SIZE + SJIS_KANJI_SIZE
#define EDIT_ERROR_CODE -4
#define EUC_KANJI_SIZE 2
#define EUC_KANJI_SIZE_1 3
#define FILE_ERROR_CODE -2;
#define FILE_NAME_SIZE 255
#define FIRST_KANJI 159
#define IDENTICAL_TIME "-a"
#define KANA_HANDAKUTEN 0xDF
#define KANA_KUTEN 0xA1
#define KATAKANA_SIZE 1
#define KANJI_SIZE 2
#define LF "-LF"
#define LINE_FEED "-L"
#define MS932_KANJI_SIZE 2
#define MS932_SJIS_BIN "MS932~Shift_JIS.bin"
#define NO_EDITING "-u"
#define NO_READING "-q"
#define NON_EUC_SEARCH 3
#define NORMAL_ARRAY_SIZE 3
#define NUMBER "-n"
#define PDF_FILE_NAME "SideAceVCL.pdf"
#define RECOGNIZE_FULL_NAME "-f"
#define RECORD_COUNT 7882
#define RIGHT_ARROW "" // Shift_JIS
#define SEARCH_COUNT 4
#define SEE_PDF "-p"
#define SIDE_ACE_TEXT "SideAce.txt"
#define SJIS_EUC_BIN "MS932~EUC-JP.bin"
#define SJIS_KANJI_SIZE 2
#define SJIS_KANJI_SIZE_1 3
#define SJIS_UTF8_BIN "MS932~UTF-8.bin"
#define SKIPPING "-i"
#define SMALL_BUFFER_SIZE 16
#define STANDARD_BUFFER_SIZE UNBROKEN_BUFFER_SIZE // Dividing.h
#define STOP_CODE -3
#define SUBSTITUTE_ARRAY_SIZE 3
#define TEMPORARY_BINARY "binary.tmp"
#define TEMPORARY_BINARY_FILE "-b"
#define UTF8_KANJI_SIZE 3
#define UTF8_KANJI_SIZE_1 4
#define VERSION_TEXT "version.txt"
#define WHICH_TABLE "-w"
#define ZENKAKU_KUHAKU 0x8140
#define ZERO "0"

#include <math.h> // floor()
#include <mem.h> // memset()
#include <stdio.h> // FILE, remove()
#include <stdlib.h> // getenv()
#include "dividing.h"
#include "Numbered.h"

enum ENCODING_METHOD {
    TO_ASCII,
    TO_EUC,
    TO_SJIS,
    TO_UTF8,
    METHOD_COUNT
};

typedef bool (*OPEN_CLOSE_FUNCTION)(char *buffer, int buffer_size);
typedef std::vector<std::string> ITEM;
typedef std::string* LINE_POINTER;
typedef LINE_POINTER (*LINE_NUMBER_FUNCTION)(LINE_POINTER p, char *buffer, char *combo);

typedef union KANJI KANJI;
union KANJI {
    unsigned char sequence[6];
    unsigned short code[3];
    unsigned int block;
};

typedef struct NORMAL_SUBSTITUTE NORMAL_SUBSTITUTE;
struct NORMAL_SUBSTITUTE {
    bool please_check;
    std::string normal;
    std::string substitute;
};

class CommandPath {
    char *data;

public:
    CommandPath() {
        data = new char[BUFFER_SIZE];
        strcpy(data, getenv("Path") );
    }

    ~CommandPath() {
        delete[] data;
    }

    char *path_name(char *extra);
    std::string pathName(const char *extra);
};

class FinalEffect {
    bool carriage_return, document, fullName, identical_time, lineBeginsWithNumber, no_removing, no_searching, produce_binary_file, skip_control_characters, whichTable;
    int corner, manager;
    std::string binaryName, normal, substitute, text_of_substitutes_list;

public:
    bool *whichFile() {
        bool *which_file = &whichTable;
        return which_file;
    }

    bool CR_LF() const {
        return carriage_return;
    }

    bool longName() const {
        return fullName;
    }

    bool noSearching() const {
        return no_searching;
    }

    bool numbering() const {
        return lineBeginsWithNumber;
    }

    bool portableDocuemntFormat() const {
        return document;
    }

    bool removeBinaryFile() const {
        return no_removing;
    }

    bool skipping() const {
        return skip_control_characters;
    }

    bool whichBinary() const {
        return whichTable;
    }

    FinalEffect(std::string & b, std::string & n, std::string & s, std::string & t, int c, int m, bool u, bool d, bool f, bool a, bool l, bool o, bool g, bool y, bool i, bool w) {
        binaryName = b;
        normal = n;
        substitute = s;
        text_of_substitutes_list = t;
        corner = c;
        manager = m;
        carriage_return = u;
        document = d;
        fullName = f;
        identical_time = a;
        lineBeginsWithNumber = l;
        no_removing = o;
        no_searching = g;
        produce_binary_file = y;
        skip_control_characters = i;
        whichTable = w;
    }

    ~FinalEffect() {
    }

    int cornerNumber() const {
        return corner;
    }

    int encodingMethod() const {
        return manager;
    }

    std::string listName() const {
        return text_of_substitutes_list;
    }

    std::string ms932() const {
        return normal;
    }

    std::string sjis() const {
        return substitute;
    }

    std::string tableName() const {
        return binaryName;
    }

};

class Regulate {
    const int ERROR_CODE;

    bool carriage_return, no_searching, whichTable;
    char *buffer;
    char file_name[ FILE_NAME_SIZE + 1 ];
    ITEM v;

public:
    Regulate(char *f, bool c, bool n, bool w): ERROR_CODE(1) {
        strcpy(file_name, f);
        carriage_return = c;
        no_searching = n;
        whichTable = w;
        buffer = new char[BUFFER_SIZE];
    }

    Regulate(ITEM & keep_log, bool c, bool n, bool w): ERROR_CODE(1) {
        v = keep_log;
        carriage_return = c;
        no_searching = n;
        whichTable = w;
        buffer = new char[BUFFER_SIZE];
    }

    ~Regulate() {
        delete[] buffer;
    }

    int readingAnotherFileToProduceEUC(char *new_name);
    int readingToProduceEUC(char *new_name);
    int readingToProduceUTF8(char *new_name);
    int skippingControlCharactersToProduceEUC(char *new_name);
    int skippingControlCharactersToProduceSJIS(char *new_name);
    int skippingControlCharactersToProduceUTF8(char *new_name);
};

class SparseSJIS {
    char *sparse;
    int size;

public:
    SparseSJIS() {
        size = 0xFC4C * SJIS_KANJI_SIZE_1;
        sparse = new char[size];
    }

    ~SparseSJIS() {
        delete[] sparse;
    }

    void clear() {
        memset(sparse, 0, size);
    }

    void populated(int index, int place, char c) {
        int k = index * SJIS_KANJI_SIZE_1 + place;
        *(sparse + k) = c;
    }

    bool setting(bool no_searching, bool *whichTable);
    std::string working(int index);
    unsigned char working(int index, int place);
};

class SparseEUC {
    char *sparse;
    int size;

public:
    SparseEUC() {
        size = 0xFC4C * EUC_KANJI_SIZE_1;
        sparse = new char[size];
    }

    ~SparseEUC() {
        delete[] sparse;
    }

    void clear() {
        memset(sparse, 0, size);
    }

    void populated(int index, int place, char c) {
        int k = index * EUC_KANJI_SIZE_1 + place;
        *(sparse + k) = c;
    }

    bool setting(bool no_searching, bool *whichTable);
    std::string working(int index);
    unsigned char working(int index, int place);
};

class SparseUTF8 {
    char *sparse;
    int size;

public:
    SparseUTF8() {
        size = 0xFC4C * UTF8_KANJI_SIZE_1;
        sparse = new char[size];
    }

    ~SparseUTF8() {
        delete[] sparse;
    }

    void clear() {
        memset(sparse, 0, size);
    }

    void populated(int index, int place, char c) {
        int k = index * UTF8_KANJI_SIZE_1 + place;
        *(sparse + k) = c;
    }

    bool setting(bool no_searching, bool *whichTable);
    std::string working(int index);
    unsigned char working(int index, int place);
};

bool binary_file_in_current_directory(char *buffer, int buffer_size);
bool binary_file_in_installed_directory(char *buffer, int buffer_size);
bool binary_file_in_path_directory(char *buffer, int buffer_size);
bool binary_file_in_specified_directory(char *buffer, int buffer_size);
bool editManyRecords(std::string & text, std::string & binaryName, bool no_searching, bool *whichTable);
bool editRecord(std::string & normal, std::string & substitute, std::string & binaryName, bool no_searching, bool *whichTable);
bool ms932_sjis_file_in_current_directory(char *buffer, int buffer_size);
bool ms932_sjis_file_in_installed_directory(char *buffer, int buffer_size);
bool ms932_sjis_file_in_path_directory(char *buffer, int buffer_size);
bool openClose(char *file_name);
bool openClose(char *file_name, bool write);
bool openClose(std::string & fileName);
bool sjis_utf8_file_in_current_directory(char *buffer, int buffer_size);
bool sjis_utf8_file_in_installed_directory(char *buffer, int buffer_size);
bool sjis_utf8_file_in_path_directory(char *buffer, int buffer_size);
int checkParametersBeforeProducingWindow();
int inverted(char *c);
int kanjiCode(std::string & line);
int produceBinaryFile(const char *name, bool no_removing);
int renameBinaryFile(const char *name, bool no_removing);
LINE_POINTER lineBeginsWithNumber(LINE_POINTER p, char *buffer, char *os);
LINE_POINTER nonnumber(LINE_POINTER p, char *buffer, char *os);
void ACE_should_disappear();
void acquire_parameters_from_text(const char *parameters_text, char *binary_name, char *normal, char *substitute, char *text_of_substitutes_list, int *corner, int *manager,
 bool *identical_time, bool *produce_binary_file, bool *carriage_return, bool *no_searching, bool *full_name, bool *skip_control_characters, bool *line_begins_with_number, bool *document, bool *no_reading,
 bool *edit_table, bool *which_table, bool *no_removing);
void binary_files_are_accessible_to_SideAce();
void versionGuide(double v, bool comments);

#endif // KANJI_UNION_H
